// multi option valuation double revisited optimized

// number of terms (of the sum) : 380
// number of options: 256
// cos method boundaries: -10, 10

#include <cutil_inline.h>
#include <cmath>

#include "CosPricer_kernel.cu"

typedef struct {
	int noo;
	double s0;
	const int* mat;
	const double* str;
} HestonCosOptionParams;

typedef struct {
	double v0;
	double sigma;
	double kappa;
	double rho;
	double theta;
} HestonCosModelParams;

double chi(int s, double c, double d) {
	double h = s*PI/(B-A);
	double t1 = h*(d-A);
	double t2 = h*(c-A);
	return 1./(1.+h*h)*(cos(t1)*exp(d)-cos(t2)*exp(c)+h*sin(t1)*exp(d)-h*sin(t2)*exp(c));
}

double phi(int s, double c, double d) {
	double h = s*PI/(B-A);
	if(s==0)
		return (d-c);
	else
		return 1./h*(sin(h*(d-A))-sin(h*(c-A)));
}


int COSPricerGPU(
		const double s0,
		const int numberofoptions, const double* strikes, const int* maturities,
		const double r, const double rf,
		const double* hestonparams,
		double* prices)
{
	// Setup
	HestonCosOptionParams model;
	model.noo = numberofoptions;
	model.s0 = s0;
	model.mat = maturities;
	model.str = strikes;

	TOptionData* h_OptionData = new TOptionData[model.noo];
	for(int i = 0; i<model.noo; i++) {
		h_OptionData[i].x = (double)log(model.s0/model.str[i]);
		h_OptionData[i].t = (double)model.mat[i]/365.;
		h_OptionData[i].r = (double)(r-rf);
		h_OptionData[i].rd = (double)r;
		h_OptionData[i].str = (double)model.str[i];
	}

	TModelData h_ModelData;
	h_ModelData.v0 =    hestonparams[4];
	h_ModelData.theta = hestonparams[1];
	h_ModelData.sigma = hestonparams[2];
	h_ModelData.kappa = hestonparams[0];
	h_ModelData.rho =   hestonparams[3];

	//CUDA_SAFE_CALL( cudaThreadSynchronize() );

	// prevaluation Uk
	double* h_Uk = new double[MAX_ITER];
	for (int i=0; i<MAX_ITER; i++) {
		h_Uk[i] = (2.f/(B-A))*(chi(i,0.f,B)-phi(i,0.f,B));//CALL
	}

	CUDA_SAFE_CALL( cudaMemcpyToSymbol(d_Uk, h_Uk, MAX_ITER * sizeof(double)) );
	CUDA_SAFE_CALL( cudaMemcpyToSymbol(d_ModelData, &h_ModelData, sizeof(TModelData)) );

	TOptionData* d_OptionData;
	CUDA_SAFE_CALL( cudaMalloc( (void**) &d_OptionData, sizeof(TOptionData)*model.noo));
	CUDA_SAFE_CALL( cudaMemcpy(d_OptionData, h_OptionData, sizeof(TOptionData)*model.noo, cudaMemcpyHostToDevice ) );
	
	double* d_Val;
	CUDA_SAFE_CALL(cudaMalloc( (void**) &d_Val, sizeof(double)*model.noo));

	int nblocks = model.noo;
	int nthreads = MAX_ITER;
	COS<<< nblocks, nthreads >>>(d_OptionData, d_Val, model.noo);

	double* h_Val = new double[model.noo];
	CUDA_SAFE_CALL( cudaMemcpy(h_Val, d_Val, sizeof(double)*model.noo, cudaMemcpyDeviceToHost) );

	for (int i=0; i<model.noo; i++) {
		prices[i] = h_Val[i];
	}

	delete [] h_Uk;
	delete [] h_Val;
	delete [] h_OptionData;
	CUDA_SAFE_CALL( cudaFree(d_OptionData) );
	CUDA_SAFE_CALL( cudaFree(d_Val) );

	return 1;
}
