#define double float
#define MAX_ITER 512

typedef struct {
	double x;
	double t;
	double r;
	double str;
	double rd;
} TOptionData;

typedef struct {
	double v0;
	double sigma;
	double kappa;
	double rho;
	double theta;
} TModelData;

static __device__ __constant__ double d_Uk[MAX_ITER];
static __device__ __constant__ TModelData d_ModelData;

typedef struct __builtin_align__(16) {
	double r,i;
} fcomplex;

#define MUL(a, b) __umul24(a, b)

#define PI (3.141592653589793)
#define B (10.)
#define A (-10.)

static __device__ __host__ inline fcomplex Cadd(fcomplex a, fcomplex b)
{
	fcomplex c;
	c.r=a.r+b.r;
	c.i=a.i+b.i;
	return c;
}

static __device__ __host__ inline fcomplex Csub(fcomplex a, fcomplex b)
{
	fcomplex c;
	c.r=a.r-b.r;
	c.i=a.i-b.i;
	return c;
}


static __device__ __host__ inline fcomplex Cmul(fcomplex a, fcomplex b)
{
	fcomplex c;
	c.r=a.r*b.r-a.i*b.i;
	c.i=a.i*b.r+a.r*b.i;
	return c;
}

static __device__ __host__ inline fcomplex Complex(double re, double im)
{
	fcomplex c;
	c.r=re;
	c.i=im;
	return c;
}

static __device__ __host__ inline fcomplex Conjg(fcomplex z)
{
	fcomplex c;
	c.r=z.r;
	c.i = -z.i;
	return c;
}

static __device__ __host__ inline fcomplex Cdiv(fcomplex a, fcomplex b)
{
	fcomplex c;
	double r,den;
	if (fabs(b.r) >= fabs(b.i)) {
		r=b.i/b.r;
		den=b.r+r*b.i;
		c.r=(a.r+r*a.i)/den;
		c.i=(a.i-r*a.r)/den;
	} else {
		r=b.r/b.i;
		den=b.i+r*b.r;
		c.r=(a.r*r+a.i)/den;
		c.i=(a.i*r-a.r)/den;
	}
	return c;
}

static __device__ __host__ inline double Cabs(fcomplex z)
{
	double x,y,ans,temp;
	x=fabs(z.r);
	y=fabs(z.i);
	if (x == 0.f)
		ans=y;
	else if (y == 0.f)
		ans=x;
	else if (x > y) {
		temp=y/x;
		ans=x*sqrt(1.f+temp*temp);
	} else {
		temp=x/y;
		ans=y*sqrt(1.f+temp*temp);
	}
	return ans;
}

static __device__ __host__ inline fcomplex Csqrt(fcomplex z)
{
	fcomplex c;
	double x,y,w,r;
	if ((z.r == 0.f) && (z.i == 0.f)) {
		c.r=0.f;
		c.i=0.f;
		return c;
	} else {
		x=fabs(z.r);
		y=fabs(z.i);
		if (x >= y) {
			r=y/x;
			w=sqrt(x)*sqrt(0.5*(1.f+sqrt(1.f+r*r)));
		} else {
			r=x/y;
			w=sqrt(y)*sqrt(0.5*(r+sqrt(1.f+r*r)));
		}
		if (z.r >= 0.f) {
			c.r=w;
			c.i=z.i/(2.f*w);
		} else {
			c.i=(z.i >= 0) ? w : -w;
			c.r=z.i/(2.f*c.i);
		}
		return c;
	}
}

static __device__ __host__ inline fcomplex Cexp(fcomplex z)
{
	double tmp = exp(z.r);
	fcomplex c;
	c.r = tmp*cos(z.i);
	c.i = tmp*sin(z.i);
	return c;
}

static __device__ __host__ inline fcomplex Clog(fcomplex z)
{
	fcomplex c;
	c.r = log(sqrt(z.r*z.r+z.i*z.i));
	c.i = atan(z.i/z.r);
	if (z.r<0 && z.i>=0)
		c.i += 3.141592653589793;
	if (z.r<0 && z.i<0)
		c.i -= 3.141592653589793;
	return c;
}

static __device__ __host__ inline fcomplex RCmul(double x, fcomplex a)
{
	fcomplex c;
	c.r=x*a.r;
	c.i=x*a.i;
	return c;
}

static __device__ inline fcomplex Init(double x, double y)
{
	fcomplex c;
	c.r=x;
	c.i=y;
	return c;
}

static __device__ inline fcomplex phiHESTON(double omega, double u_0, double sigma, double kappa, double rho, double theta, double t, double r) {
	
	fcomplex tmp;
	tmp.r=0.;
	tmp.i=1.;
	tmp = RCmul(rho*sigma*omega, tmp);
	fcomplex h4;
	h4.r=kappa-tmp.r;
	h4.i=-tmp.i;
	tmp.r=sigma*sigma*omega*omega;
	tmp.i=sigma*sigma*omega;
	fcomplex D=Csqrt(Cadd(Cmul(h4,h4), tmp));
	fcomplex h2 = Csub(h4,D);
	fcomplex h3 = Cadd(h4,D);
	fcomplex G = Cdiv(h2,h3);
	//cout << "h4: " << h4.r << " " << h4.i << " G: " << G.r << " " << G.i << std::endl;
	fcomplex h5;
	h5.r = 0.;
	h5.i = (omega*r*t);
	fcomplex h6a= RCmul(-t,D);
	fcomplex h6;
	h6.r = exp(h6a.r)*cos(h6a.i);
	h6.i = exp(h6a.r)*sin(h6a.i);
	fcomplex h7;h7.r = 1.-Cmul(G,h6).r;h7.i = -Cmul(G,h6).i;
	fcomplex h8;h8.r = 1.-h6.r;h8.i = -h6.i;
	fcomplex h9 =Cdiv(h8,h7);
	fcomplex h10 = Cadd(h5,RCmul(u_0/(sigma*sigma),Cmul(h9,h2)));
	fcomplex h11;
	h11.r=1.-G.r;
	h11.i=-G.i;

	fcomplex val=Cmul(Cexp(h10),Cexp(RCmul(kappa*theta/(sigma*sigma),Csub(RCmul(t,h2),RCmul(2.f,Clog(Cdiv(h7,h11)))))));
	/*cout<<"phiheston:"<<endl;
	cout<<val.r<<" "<<val.i<<endl;*/
	return val;
}

__global__ void COS(const TOptionData* d_OptionData, double* d_Val, int noo) {

	__shared__ double tmp[MAX_ITER];
	__shared__ TOptionData blkOptionData;

	if (threadIdx.x == 0) {
		blkOptionData = d_OptionData[blockIdx.x];
	}

	__syncthreads();

	fcomplex im;
	im.r=0.;
	im.i=threadIdx.x*PI*(blkOptionData.x-A)/(B-A);

	tmp[threadIdx.x] = Cmul(
		phiHESTON(threadIdx.x*PI/(B-A),
		d_ModelData.v0,
		d_ModelData.sigma,
		d_ModelData.kappa,
		d_ModelData.rho,
		d_ModelData.theta,
		blkOptionData.t,
		blkOptionData.r),RCmul(d_Uk[threadIdx.x],Cexp(im))).r;

	__syncthreads();

	if (threadIdx.x == MAX_ITER - 1) {
		tmp[0] *= 0.5;	
		for (int i = MAX_ITER - 2; i>=0; i--) {
			tmp[MAX_ITER - 1] += tmp[i];
		}
<<<<<<< .mine
		//tmp[511] = max(tmp[511],0.);	
		d_Val[blockIdx.x] = d_OptionData[blockIdx.x].str*exp(-d_OptionData[blockIdx.x].rd*d_OptionData[blockIdx.x].t)*tmp[511];
=======
		tmp[MAX_ITER - 1] = max(tmp[MAX_ITER - 1],0.);	
		d_Val[blockIdx.x] = blkOptionData.str*exp(-blkOptionData.rd*blkOptionData.t)*tmp[MAX_ITER - 1];
>>>>>>> .r3121
	}

//	d_Val[blockIdx.x] = -1.f;
}
